package asmc

import (
	"encoding/binary"
	"fmt"

	"git.urbach.dev/cli/q/src/arm"
	"git.urbach.dev/cli/q/src/asm"
)

type armCompiler struct {
	*compiler
}

func (c *armCompiler) Compile(instruction asm.Instruction) {
	switch instruction.Mnemonic {
	case asm.MOVE:
		c.handleMoveInstruction(instruction)
	case asm.CALL:
		c.handleCallInstruction(instruction)
	case asm.LABEL:
		c.handleLabelInstruction(instruction)
	case asm.LOAD:
		c.handleLoadInstruction(instruction)
	case asm.STORE:
		c.handleStoreInstruction(instruction)
	case asm.RETURN:
		c.append(arm.LoadPair(arm.FP, arm.LR, arm.SP, 16))
		c.append(arm.Return())
	case asm.SYSCALL:
		c.append(arm.Syscall())
	case asm.PUSH:
		c.handlePushInstruction(instruction)
	case asm.POP:
		c.handlePopInstruction(instruction)
	case asm.AND:
		c.handleAndInstruction(instruction)
	case asm.OR:
		c.handleOrInstruction(instruction)
	case asm.XOR:
		c.handleXorInstruction(instruction)
	case asm.ADD:
		c.handleAddInstruction(instruction)
	case asm.SUB:
		c.handleSubInstruction(instruction)
	case asm.COMPARE:
		c.handleCompareInstruction(instruction)
	case asm.DIV:
		c.handleDivInstruction(instruction)
	case asm.MUL:
		c.handleMulInstruction(instruction)
	case asm.MODULO:
		c.handleModuloInstruction(instruction)
	case asm.JE, asm.JNE, asm.JG, asm.JGE, asm.JL, asm.JLE, asm.JUMP:
		c.handleJumpInstruction(instruction)
	case asm.SHIFTL:
		c.handleShiftLeftInstruction(instruction)
	case asm.SHIFTRS:
		c.handleShiftRightSignedInstruction(instruction)
	case asm.NEGATE:
		c.handleNegateInstruction(instruction)
	default:
		panic("unknown mnemonic: " + instruction.Mnemonic.String())
	}
}

func (c *armCompiler) handleMoveInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterRegister:
		operands := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.MoveRegisterRegister(operands.Destination, operands.Source))
	case asm.TypeRegisterNumber:
		operands := c.assembler.Param.RegisterNumber[instruction.Index]
		c.code = arm.MoveRegisterNumber(c.code, operands.Register, operands.Number)
	case asm.TypeRegisterLabel:
		operands := c.assembler.Param.RegisterLabel[instruction.Index]
		position := Address(len(c.code))
		c.append(arm.LoadAddress(operands.Register, 0))

		if operands.Label.Type == asm.DataLabel {
			c.dataPointers = append(c.dataPointers, c.createDataPointer(operands, position))
		} else {
			c.codePointers = append(c.codePointers, c.createCodePointer(operands, position))
		}
	}
}

func (c *armCompiler) handleCallInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeLabel:
		label := c.assembler.Param.Label[instruction.Index]
		position := Address(len(c.code))
		c.append(arm.Call(0))

		pointer := &pointer{
			Position: position,
			OpSize:   0,
			Size:     4,
		}

		pointer.Resolve = func() Address {
			destination, exists := c.codeLabels[label.Name]

			if !exists {
				panic(fmt.Sprintf("unknown jump label %s", label.Name))
			}

			distance := int(destination-position) / 4
			return arm.Call(distance)
		}

		c.codePointers = append(c.codePointers, pointer)
	default:
		panic("not implemented")
	}
}

func (c *armCompiler) handleLabelInstruction(instruction asm.Instruction) {
	label := c.assembler.Param.Label[instruction.Index]
	c.codeLabels[label.Name] = Address(len(c.code))

	if label.Type == asm.FunctionLabel {
		c.append(arm.StorePair(arm.FP, arm.LR, arm.SP, -16))
		c.append(arm.MoveRegisterRegister(arm.FP, arm.SP))
	}
}

func (c *armCompiler) handleLoadInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeMemoryRegister:
		operands := c.assembler.Param.MemoryRegister[instruction.Index]

		if operands.Address.OffsetRegister < 0 {
			c.append(arm.LoadRegister(operands.Register, operands.Address.Base, int(operands.Address.Offset), operands.Address.Length))
		} else {
			panic("not implemented")
		}
	}
}

func (c *armCompiler) handleStoreInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeMemoryRegister:
		operands := c.assembler.Param.MemoryRegister[instruction.Index]

		if operands.Address.OffsetRegister < 0 {
			c.append(arm.StoreRegister(operands.Register, operands.Address.Base, int(operands.Address.Offset), operands.Address.Length))
		} else {
			panic("not implemented")
		}
	case asm.TypeMemoryNumber:
		operands := c.assembler.Param.MemoryNumber[instruction.Index]

		if operands.Address.OffsetRegister < 0 {
			c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operands.Number)
			c.append(arm.StoreRegister(arm.TMP, operands.Address.Base, int(operands.Address.Offset), operands.Address.Length))
		} else {
			panic("not implemented")
		}
	}
}

func (c *armCompiler) handlePushInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegister:
		operand := c.assembler.Param.Register[instruction.Index]
		code, _ := arm.SubRegisterNumber(arm.SP, arm.SP, 16)
		c.append(code)
		c.append(arm.StoreRegister(operand.Register, arm.SP, 0, 8))
	}
}

func (c *armCompiler) handlePopInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegister:
		operand := c.assembler.Param.Register[instruction.Index]
		c.append(arm.LoadRegister(operand.Register, arm.SP, 0, 8))
		code, _ := arm.AddRegisterNumber(arm.SP, arm.SP, 16)
		c.append(code)
	}
}

func (c *armCompiler) handleAndInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		code, encodable := arm.AndRegisterNumber(operand.Register, operand.Register, operand.Number)

		if encodable {
			c.append(code)
		} else {
			c.code = arm.MoveRegisterNumberMI(c.code, arm.TMP, operand.Number)
			c.append(arm.AndRegisterRegister(operand.Register, operand.Register, arm.TMP))
		}
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.AndRegisterRegister(operand.Destination, operand.Destination, operand.Source))
	}
}

func (c *armCompiler) handleOrInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		code, encodable := arm.OrRegisterNumber(operand.Register, operand.Register, operand.Number)

		if encodable {
			c.append(code)
		} else {
			c.code = arm.MoveRegisterNumberMI(c.code, arm.TMP, operand.Number)
			c.append(arm.OrRegisterRegister(operand.Register, operand.Register, arm.TMP))
		}
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.OrRegisterRegister(operand.Destination, operand.Destination, operand.Source))
	}
}

func (c *armCompiler) handleXorInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		code, encodable := arm.XorRegisterNumber(operand.Register, operand.Register, operand.Number)

		if encodable {
			c.append(code)
		} else {
			c.code = arm.MoveRegisterNumberMI(c.code, arm.TMP, operand.Number)
			c.append(arm.XorRegisterRegister(operand.Register, operand.Register, arm.TMP))
		}
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.XorRegisterRegister(operand.Destination, operand.Destination, operand.Source))
	}
}

func (c *armCompiler) handleAddInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		code, encodable := arm.AddRegisterNumber(operand.Register, operand.Register, operand.Number)

		if encodable {
			c.append(code)
		} else {
			c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operand.Number)
			c.append(arm.AddRegisterRegister(operand.Register, operand.Register, arm.TMP))
		}
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.AddRegisterRegister(operand.Destination, operand.Destination, operand.Source))
	}
}

func (c *armCompiler) handleSubInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		code, encodable := arm.SubRegisterNumber(operand.Register, operand.Register, operand.Number)

		if encodable {
			c.append(code)
		} else {
			c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operand.Number)
			c.append(arm.SubRegisterRegister(operand.Register, operand.Register, arm.TMP))
		}
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.SubRegisterRegister(operand.Destination, operand.Destination, operand.Source))
	}
}

func (c *armCompiler) handleCompareInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		code, encodable := arm.CompareRegisterNumber(operand.Register, operand.Number)

		if encodable {
			c.append(code)
		} else {
			c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operand.Number)
			c.append(arm.CompareRegisterRegister(operand.Register, arm.TMP))
		}
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.CompareRegisterRegister(operand.Destination, operand.Source))
	}
}

func (c *armCompiler) handleDivInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.DivSigned(operand.Destination, operand.Destination, operand.Source))
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operand.Number)
		c.append(arm.DivSigned(operand.Register, operand.Register, arm.TMP))
	}
}

func (c *armCompiler) handleMulInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.MulRegisterRegister(operand.Destination, operand.Destination, operand.Source))
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operand.Number)
		c.append(arm.MulRegisterRegister(operand.Register, operand.Register, arm.TMP))
	}
}

func (c *armCompiler) handleModuloInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterRegister:
		operand := c.assembler.Param.RegisterRegister[instruction.Index]
		c.append(arm.DivSigned(arm.TMP, operand.Destination, operand.Source))
		c.append(arm.MultiplySubtract(operand.Destination, arm.TMP, operand.Source, operand.Destination))
	case asm.TypeRegisterNumber:
		operand := c.assembler.Param.RegisterNumber[instruction.Index]
		c.code = arm.MoveRegisterNumber(c.code, arm.TMP, operand.Number)
		c.append(arm.DivSigned(arm.TMP2, operand.Register, arm.TMP))
		c.append(arm.MultiplySubtract(operand.Register, arm.TMP2, arm.TMP, operand.Register))
	}
}

func (c *armCompiler) handleJumpInstruction(x asm.Instruction) {
	label := c.assembler.Param.Label[x.Index]
	mnemonic := x.Mnemonic
	position := Address(len(c.code))

	pointer := &pointer{
		Position: position,
		Size:     4,
	}

	c.append(arm.Nop())

	pointer.Resolve = func() Address {
		destination, exists := c.codeLabels[label.Name]

		if !exists {
			panic(fmt.Sprintf("unknown jump label %s", label.Name))
		}

		distance := (int(destination) - int(position)) / 4

		switch mnemonic {
		case asm.JE:
			return arm.JumpIfEqual(distance)
		case asm.JNE:
			return arm.JumpIfNotEqual(distance)
		case asm.JG:
			return arm.JumpIfGreater(distance)
		case asm.JGE:
			return arm.JumpIfGreaterOrEqual(distance)
		case asm.JL:
			return arm.JumpIfLess(distance)
		case asm.JLE:
			return arm.JumpIfLessOrEqual(distance)
		case asm.JUMP:
			return arm.Jump(distance)
		default:
			panic("not implemented")
		}
	}

	c.codePointers = append(c.codePointers, pointer)
}

func (c *armCompiler) handleShiftLeftInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operands := c.assembler.Param.RegisterNumber[instruction.Index]
		c.append(arm.ShiftLeftNumber(operands.Register, operands.Register, operands.Number&0b111111))
	case asm.TypeRegisterRegister:
		panic("not implemented")
	}
}

func (c *armCompiler) handleShiftRightSignedInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegisterNumber:
		operands := c.assembler.Param.RegisterNumber[instruction.Index]
		c.append(arm.ShiftRightSignedNumber(operands.Register, operands.Register, operands.Number&0b111111))
	case asm.TypeRegisterRegister:
		panic("not implemented")
	}
}

func (c *armCompiler) handleNegateInstruction(instruction asm.Instruction) {
	switch instruction.Type {
	case asm.TypeRegister:
		operands := c.assembler.Param.Register[instruction.Index]
		c.append(arm.NegateRegister(operands.Register, operands.Register))
	}
}

func (c *armCompiler) createCodePointer(operands asm.RegisterLabel, position Address) *pointer {
	return &pointer{
		Position: position,
		OpSize:   0,
		Size:     4,
		Resolve: func() Address {
			destination, exists := c.codeLabels[operands.Label.Name]

			if !exists {
				panic("unknown label")
			}

			distance := destination - position + 8
			return arm.LoadAddress(operands.Register, int(distance))
		},
	}
}

func (c *armCompiler) createDataPointer(operands asm.RegisterLabel, position Address) *pointer {
	return &pointer{
		Position: position,
		OpSize:   0,
		Size:     4,
		Resolve: func() Address {
			destination, exists := c.dataLabels[operands.Label.Name]

			if !exists {
				panic("unknown label")
			}

			destination += c.dataStart - c.codeStart
			distance := destination - position + 8
			return arm.LoadAddress(operands.Register, int(distance))
		},
	}
}

func (c *compiler) append(code uint32) {
	c.code = binary.LittleEndian.AppendUint32(c.code, code)
}
