package core

import (
	"fmt"

	"git.urbach.dev/cli/q/src/asm"
	"git.urbach.dev/cli/q/src/ast"
	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/eval"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// CompileAssign compiles an assign statement.
func (f *Function) CompileAssign(node *ast.Assign) error {
	left := node.Expression.Children[0]
	right := node.Expression.Children[1]

	if left.Token.Kind == token.Separator {
		if right.Token.Kind == token.Div {
			return f.CompileAssignDivision(node.Expression)
		}

		if right.Token.Kind == token.Call {
			return f.MultiAssign(left, right)
		}

		return errors.New(errors.NotImplemented, f.File, node.Expression.Token.Position)
	}

	leftValue, err := f.Evaluate(left)

	if err != nil {
		return err
	}

	operation := node.Expression.Token

	switch leftValue := leftValue.(type) {
	case *eval.Register:
		if !f.RegisterIsUsed(leftValue.Register) {
			f.UseRegister(leftValue.Register)
			defer f.FreeRegister(leftValue.Register)
		}

		f.Execute(operation, leftValue.Register, right)
	case *eval.Memory:
		if !f.RegisterIsUsed(leftValue.Memory.Base) {
			f.UseRegister(leftValue.Memory.Base)
			defer f.FreeRegister(leftValue.Memory.Base)
		}

		if operation.Kind == token.Assign {
			rightValue, err := f.Evaluate(right)

			if err != nil {
				return err
			}

			if !types.Is(rightValue.Type(), leftValue.Typ) {
				return errors.New(&errors.TypeMismatch{Encountered: rightValue.Type().Name(), Expected: leftValue.Typ.Name()}, f.File, right.Token.Position)
			}

			leftSize := leftValue.Memory.Length
			rightSize := uint8(rightValue.Type().Size())

			switch {
			case leftSize == 0:
				leftSize = rightSize
				leftValue.Memory.Length = rightSize
			case rightSize == 0:
				rightSize = leftSize
			}

			if leftSize == 0 {
				return errors.New(errors.UnknownMemorySize, f.File, operation.Position)
			}

			f.ValueToMemory(rightValue, leftValue.Memory)
			return nil
		}

		tmp := f.NewRegister()
		f.ValueToRegister(leftValue, tmp)
		err := f.Execute(operation, tmp, right)

		if err != nil {
			return err
		}

		f.MemoryRegister(asm.STORE, leftValue.Memory, tmp)
		f.FreeRegister(tmp)
	default:
		panic(fmt.Errorf("%s: not implemented: %v", f.UniqueName, leftValue))
	}

	return nil
}
